inlets = 1;
outlets = 1;

var i_len = 144;
var i_stride = 2;
var lowest_note = 23;

var i_bgcolor = [0, 0, 0, 0];
var i_hcolor = [0, 255, 255, 255];
var i_lcolor = [0, 0, 255, 0];

var StateColors = [
	[0,   0,  64,   0],
	[0, 128, 128,   0],
	[0, 128,   0,   0],
	[0, 255, 255, 255],
    [0,   0,   0, 255] ];
	
var State = Object.freeze({
	HARMONIC: 0,
	MAYBE: 1,
	INHARMONIC: 2,
    INUSE: 3,
    UNKNOWN: 4 });

function StateRange(_min_key, _max_key) {
	this.min_key = _min_key;
	this.max_key = _max_key;
	this.len = _max_key - _min_key + 1;
	this.arr = Array.apply(null, new Array(this.len)).map(
		function(x){return State.HARMONIC;} );
}

// Range check the key and only "upgrade" the state 
StateRange.prototype.SetState = function(key, state) {
	key -= this.min_key;
	if( key >= 0 && key < this.len && this.arr[key] <= state ) {
		this.arr[key] = state;
	}

}

// Range check the key and only "upgrade" the state 
StateRange.prototype.GetState = function(key) {
	key -= this.min_key;
	if( key >= 0 && key < this.len ) {
		return this.arr[key];
	}
	return State.UNKNOWN;
}

// Range check the key and only "upgrade" the state 
StateRange.prototype.GetStateColor = function(key) {
	return StateColors[this.GetState(key)];
}



/*
var i_bgcolor = [0, 0, 0, 0];
var i_hcolor = [0, 6, 4, 4];
var i_lcolor = [0, 0, 2, 0];
*/
//var i_bgcolor = [0, 0, 0, 0];
//var i_hcolor = [0, 255, 255, 255];
//var i_lcolor = [0, 255, 0, 0];

var mtrx = new JitterMatrix(4, "char", i_len, 1 );

function bang() {
	// output processed matrix into Max
	outlet(0, "jit_matrix", mtrx.name);		
}

function bgcolor(r,g,b) {
	i_bgcolor = [255,r,g,b];
}

function hcolor(r,g,b) {
	i_hcolor = [255,r,g,b];
}

function lcolor(r,g,b) {
	i_lcolor = [255,r,g,b];
}

var NPC = {
	Unknown : -1,
	PitchOff : 0,
	Pressure : 1,
	Mod : 2,
	Key: 3
};

var InputMode = {
	RAW : 0,
	HAKEN : 1
};

var MidiClass = {
    UNKNOWN : 0,
    SYSTEM_REALTIME : 1,
    SYSTEM_COMMON : 2,
    VOICE : 3
};

function DataMapper(_min, _max, curve) {
	this.in_min = this.out_min = _min || 0;
	this.in_max = this.out_max = _max || 127;
	this.curve = curve || 1;
}

DataMapper.prototype.SetInMin = function(in_min) {
	this.in_min = in_min;
}

DataMapper.prototype.SetInMax = function(in_max) {
	this.in_max = in_max;
}

DataMapper.prototype.SetOutMin = function(out_min) {
	this.out_min = out_min;
}

DataMapper.prototype.SetOutMax = function(out_max) {
	this.out_max = out_max;
}

DataMapper.prototype.SetCurve = function(curve) {
	this.curve = curve;
}

DataMapper.prototype.Map = function(val) {
	return Math.round( ( Math.pow( ( ( val - this.in_min ) / (this.in_max - this.in_min) ), this.curve ) *
		(this.out_max - this.out_min) ) + this.out_min );
}

function MidiMsg( msg ) {
    this.msg = msg;
}

MidiMsg.prototype.Class = function() {
    if( this.msg[0] >= 0xF8 ) {
        return MidiClass.SYSTEM_REALTIME;
    } else if( this.msg[0] >= 0xF0 ) {
        return MidiClass.SYSTEM_COMMON;
    } else if( this.msg[0] >= 0x80 ) {
        return MidiClass.VOICE;
    } else {
        return MidiClass.UNKNOWN;
    }
}

MidiMsg.prototype.Channel = function() {
    if( this.Class() == MidiClass.VOICE ) {
        return this.msg[0] & 0xf;
    }
    return null;
}

MidiMsg.prototype.SetChannel = function(chan) {
    if( this.Class() == MidiClass.VOICE ) {
        this.msg[0] = (this.msg[0] & 0xf0) | (chan & 0xf);
		return true;
    }
    return false;
}

MidiMsg.CreateController = function(ctrl, value, chan) {
	chan = chan || 0
	return new MidiMsg( [ (0xB0 | (chan & 0xf)), (ctrl & 0x7f), (value & 0x7f) ] );
}

MidiMsg.prototype.IsController = function() {
	return (this.msg[0] & 0xF0) == 0xB0;
}

MidiMsg.prototype.Controller = function() {
	if( this.IsController() ) {
		return this.msg[1];
	}
	return null;
}

MidiMsg.prototype.ControllerValue = function() {
	if( (this.msg[0] & 0xF0) == 0xB0 ) {
		return this.msg[2];
	}
	return null;
}

MidiMsg.CreatePitchwheel = function(value, chan) {
	chan = chan || 0
	value += 8192;
	return new MidiMsg( [ 0xE0 | (chan & 0xf), value & 0x7f, (value >> 7) & 0x7f ] );
}

MidiMsg.prototype.IsPitchWheel = function() {
	return (this.msg[0] & 0xF0) == 0xE0;
}

MidiMsg.prototype.PitchWheel = function() {
	if( this.IsPitchWheel() ) {
		return ( this.msg[1] | ( this.msg[2] << 7 ) ) - 8192;
	}
	return null;
}

MidiMsg.prototype.PitchWheelRaw = function() {
	if( this.IsPitchWheel() ) {
		return ( this.msg[1] | ( this.msg[2] << 7 ) );
	}
	return null;
}

MidiMsg.CreateNoteOn = function(key, vel, chan) {
	chan = chan || 0
	return new MidiMsg( [ 0x90 | (chan & 0xf), key & 0x7f, vel & 0x7f ] );
}

MidiMsg.CreateNoteOff = function(key, vel, chan) {
	chan = chan || 0
	vel = vel || 127
	return new MidiMsg( [ 0x80 | (chan & 0xf), key & 0x7f, vel & 0x7f ] );
}

MidiMsg.prototype.IsNote = function() {
	return (this.msg[0] & 0xE0) == 0x80;
}

MidiMsg.prototype.SetNoteVel = function(vel) {
	if( this.IsNote() ) {
		this.msg[2] = vel & 0x7f;
	}
}

MidiMsg.prototype.IsNoteOn = function() {
	return ((this.msg[0] & 0xF0) == 0x90) && (this.msg[2] != 0);
}

MidiMsg.prototype.IsNoteOff = function() {
	return ( (this.msg[0] & 0xF0) == 0x80 ) ||
		(((this.msg[0] & 0xF0) == 0x90) && (this.msg[2] == 0));
}

MidiMsg.prototype.NoteOnKeyVel = function() {
	if( ( (this.msg[0] & 0xF0) == 0x90 ) && (this.msg[2] != 0) ) {
		return this.msg.slice(1,3);
	}
	return [-1, -1];
}

MidiMsg.prototype.NoteOffKeyVel = function() {
	if( (this.msg[0] & 0xF0) == 0x80 ) {
		return this.msg.slice(1,3);
	} else if( ( (this.msg[0] & 0xF0) == 0x90 ) && (this.msg[2] == 0) ) {
		return [ this.msg[1], 127 ];
	}
	return [-1, -1];
}

function MidiAccumulator() {
    this._offset = 0;
    this._length = 0;
    this._msg = [];
}

MidiAccumulator.prototype._MidiSysCommonBytes = [ -1, 2, 3, 2, -1, -1, 1, -2 ];

MidiAccumulator.prototype.HandleMessage = function( msg ) {
    post( "This should be replaced with user code" );
	post();
}

MidiAccumulator.prototype.HandleError = function( msg ) { }

// Handle the high level control logic for accumulating a MIDI message
MidiAccumulator.prototype.Accumulate = function( val ) {

    // Realtime Messages
    if( val >= 0xF8 ) {
        this._NewStatus( val, 0 );

    // System Common Messages
    } else if( val >= 0xF0 ) {
        this._NewStatus( val, this._MidiSysCommonBytes[ v & 0xf ] );

    // 2-byte Voice Commands
    } else if( val >= 0xC0 && val <= 0xDF ) {
        this._NewStatus( val, 2 );

    // 3-byte Voice Commands
    } else if( val >= 0x80 ) {
        this._NewStatus( val, 3 );

    // Any other message just needs to be accumulated
    } else {
        this._AccumMultiByte( val );
    }
}

// Accumulate a byte in a multi-byte message
MidiAccumulator.prototype._AccumMultiByte = function( val ) {

    // Length -1 = In SysEx and accumulate forever 
    if( this._length == -1 ) {
        this._msg.push( val );

    // Length 2+ = Accumulate another message with the current status
    } else if( this._length > 1 ) {

        // Inject the byte into the multi-byte buffer
        this._msg[this._offset] = val;
        this._offset++;

        // If we hit the correct numer of bytes, handle the message
        if( this._offset == this._length ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._offset = 1;
        }

    // Any other length is an error
    } else {
        this.HandleError( [val] );
    }
}

// Handle the boundaries relating to a new status message
MidiAccumulator.prototype._NewStatus = function ( val, len ) {

    // Length 0 = System Realtime (doesn't effect the current status)
    if( len == 0 ) {
        this.HandleMessage( new MidiMsg( [val] ) );

    // Length -2 = End of a SysEx
    } else if( len == -2 ) {
        this._AccumMultiByte( val );
        this.HandleMessage( this._msg );
        this._msg.length = 1;

    // Handle real status changes here
    } else {
        this._msg = [val];
        this._offset = 1;
        this._length = len;

        // Length 1 = Handle immediately (no data required)
        if( len == 1 ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._length = 1;

        // Any other length means we need to accumulate messages
        } else if( len > 1 ) {
            this._msg.length = len;
        }
    }
}

// KeyTracker handles a given key on a channel on a port.
// At this level we handle:
//   Key Down
//   Key Up
//   Sustain (just a small part) 

function KeyTracker(parent, key_num) {
	this.parent = parent;
	this.key = key_num;
	this.down_count = 0;
	this.sustain_count = 0;
}

KeyTracker.prototype.count = function() {
	return this.down_count + this.sustain_count + this.parent.sustenuto;
}

KeyTracker.prototype.KeyDown = function() {
	this.down_count += 1;
}

KeyTracker.prototype.KeyUp = function() {
	this.down_count -= 1;
	if( this.parent.sustain ) {
		this.sustain_count += 1;
	}
}

KeyTracker.prototype.SustainUp = function() {
	this.sustain_count = 0;
}

KeyTracker.prototype.AdjustedKey = function() {
	return this.key + this.parent.computed_transpose;
}

// ChannelTracker handles a given channel on a port and all the resources under it.
// At this level we handle:
//   Pitch Bend
//   Sustain
//   Sustenuto
//   All Notes Off
//   Reset All Controllers 

function ChannelTracker(parent) {
	this.parent = parent;
	this.keys = {};
	this.pitchbend = 0.0;
	this.computed_transpose = 0.0;
	this.sustain = false;
	this.sustenuto = false;
}

ChannelTracker.prototype.KeyDown = function(key) {
	kobj = this.keys[key];
	if( kobj === undefined ) {
		kobj = this.keys[key] = new KeyTracker(this, key);
		this.parent.parent.MarkNeedsRefresh();
	}
	kobj.KeyDown();
}

ChannelTracker.prototype.KeyCheck = function(key) {
	if( this.keys[key].count() < 1 ) {
		delete this.keys[key];
		this.parent.parent.MarkNeedsRefresh();
	}

}

ChannelTracker.prototype.KeyUp = function(key) {
	kobj = this.keys[key];
	if( kobj !== undefined ) {
		kobj.KeyUp();
		this.KeyCheck(key);
	}
}

// Pitch Bend is assumed to be normalized to +/-1
ChannelTracker.prototype.PitchBend = function(val) {
	this.pitchbend = val;
	this._Recompute();
}

ChannelTracker.prototype._Recompute = function() {
	old_ct = this.computed_transpose;
	this.computed_transpose = this.parent.transpose + Math.round( this.pitchbend * this.parent.pitchbend_range );
	
	// If we moved by more than 1, we need to refresh
	if( Math.abs( this.computed_transpose - old_ct ) > 0.99 ) {
		this.parent.parent.MarkNeedsRefresh();
	}	
}

ChannelTracker.prototype.Controller = function(num, val) {

	// Sustain Control
	if( num == 64 ) {
		this.sustain = (val >= 64);
		if( !this.sustain ) {
			for( var key in this.keys ) {
				this.keys[key].SustainUp();
				this.KeyCheck(key);
			}
		}
		
	// Sustenuto control
	} else if( num == 66 ) {
		this.sustenuto = (val >= 64);
	
	// Reset All Controllers
	} else if( num == 121 ) {
		this.sustain = false;
		this.sustenuto = false;
		this.Pitchbend(0);
		
	// All Notes Off
	} else if( num == 123 ) {
		this.keys = {};
		this.parent.parent.MarkNeedsRefresh();
	}
}

// PortTracker handles a given port and all the resources under it.
// At this level we handle:
//   Pitch Bend Range
//   Transpose 

function PortTracker(parent) {
	this.parent = parent;
	this.pitchbend_range = 2;
	this.transpose = 0;
	
	// Sse another variable so it can be bound into the array apply function
	self = this;
	
	// Setup all the channels
	this.channels = Array.apply(null, new Array(16)).map(
		function(x){return new ChannelTracker(self);} );

	this.accum = new MidiAccumulator();
	this.accum.pt = this;
	this.accum.HandleMessage = function(msg) {
		ch = msg.Channel();
		
		if( ch !== null ) {
			if( msg.IsNoteOn() ) {
				this.pt.KeyDown( ch, msg.NoteOnKeyVel()[0] );
			} else if( msg.IsNoteOff() ) {
				this.pt.KeyUp( ch, msg.NoteOffKeyVel()[0] );
			} else if( msg.IsPitchWheel() ) {
				this.pt.PitchBend( ch, msg.PitchWheel() / 8192.0 );
			} else if( msg.IsController() ) {
				this.pt.Controller( ch, msg.Controller(), msg.ControllerValue() );
			}
		}
	}
}

PortTracker.prototype.AccumulateMidiMsg = function(midi_msg) {
	this.accum.Accumulate(midi_msg);
}

PortTracker.prototype.SetTranspose = function(val) {
	this.transpose = val;
	this._Recompute();
}

PortTracker.prototype.SetPitchBendRange = function(val) {
	this.pitchbend_range = val;
	this._Recompute();
}

PortTracker.prototype._Recompute = function() {
	for(var chan in this.channels) {
		this.channels[chan]._Recompute();
	}
	this.parent.MarkNeedsRefresh();
}

PortTracker.prototype.KeyDown = function(channel, key) {
	this.channels[channel].KeyDown(key);
}

PortTracker.prototype.KeyUp = function(channel, key) {
	this.channels[channel].KeyUp(key);
}

PortTracker.prototype.PitchBend = function(channel, val) {
	this.channels[channel].PitchBend(val);
}

PortTracker.prototype.Controller = function(channel, num, val) {
	this.channels[channel].Controller(num, val);
}

// Tracker handles the top level for all incoming MIDI ports.
// At this level we handle:
//   Port Allocation

function Tracker() {
	this.ports = {}
	this.needs_refresh = true;
}

Tracker.prototype.GetPort = function(port) {
	return this.ports[port] || (this.ports[port] = new PortTracker(this));
}

Tracker.prototype.AccumulateMidiMsg = function(port, midi_msg) {
	this.GetPort(port).AccumulateMidiMsg(midi_msg);
}

Tracker.prototype.MarkNeedsRefresh = function() {
	this.needs_refresh = true;
}

tracker = new Tracker();

tracker.update = function() {
	if( this.needs_refresh ) {
		mtrx.setall.apply(mtrx, i_bgcolor);
		vlen = i_len / i_stride;
		
		st_arr = new StateRange(lowest_note, lowest_note+vlen);		

		// Walk all keys, mapping the data
		for(var port in this.ports) {
			pobj = this.ports[port]
			for(var chan in pobj.channels) {
				cobj = pobj.channels[chan];
				for(var key in cobj.keys) {
					var adj_key = cobj.keys[key].AdjustedKey();
					st_arr.SetState(adj_key-13, State.MAYBE);
					st_arr.SetState(adj_key-11, State.MAYBE);
					st_arr.SetState(adj_key-2, State.INHARMONIC);
					st_arr.SetState(adj_key-1, State.INHARMONIC);
					st_arr.SetState(adj_key, State.INUSE);
					st_arr.SetState(adj_key+1, State.INHARMONIC);
					st_arr.SetState(adj_key+2, State.INHARMONIC);
					st_arr.SetState(adj_key+11, State.MAYBE);
					st_arr.SetState(adj_key+13, State.MAYBE);
				}
			}
		}
		
		for(var key=lowest_note, x=i_len; x >= 0; key++, x -= i_stride) {
			mtrx.setcell2d.apply(mtrx, [x, 0].concat(st_arr.GetStateColor(key)));
		}
		this.needs_refresh = false;
		bang();
	}
}

tracker.debug = function() {	
	
	// Walk all keys, mapping the data
	for(var port in this.ports) {
		pobj = this.ports[port]
		for(var chan in pobj.channels) {
			cobj = pobj.channels[chan];
			for(var key in cobj.keys) {
				var adj_key = cobj.keys[key].AdjustedKey();
				post("Port");
				post(port);
				post("Channel");
				post(chan);
				post("Key");
				post(adj_key);
				post("Count");
				post(cobj.keys[key].count());
				post();
			}
		}
	}
}

tracker.reset = function() {	
	
	// Walk all keys, mapping the data
	for(var port in this.ports) {
		pobj = this.ports[port]
		for(var chan in pobj.channels) {
			cobj = pobj.channels[chan];
			cobj.keys = {}
		}
	}
	this.MarkNeedsRefresh();
}

tracker.updateTask = new Task(tracker.update, tracker);
tracker.updateTask.interval = 50;
tracker.updateTask.repeat();

function showDebug() {
	tracker.debug();
}

function reset() {
	tracker.reset();
}

function pitchbend(port_num, range) {
	tracker.GetPort(port_num).SetPitchBendRange(range);
}

function transpose(port_num, off) {
	tracker.GetPort(port_num).SetTranspose(off);
}

function port(port_num, midi_data) {
    tracker.AccumulateMidiMsg(port_num, midi_data);
//	tracker.update();
}
